/* The code of interest is in the subroutine GetDriveGeometry. The 
   code in main shows how to interpret the results of the IOCTL call. */

#include <stdio.h>
#include <windows.h>
#include <winioctl.h>

BOOL GetDriveGeometry(DISK_GEOMETRY *pdg)
{
  HANDLE hDevice;               // handle to the drive to be examined 
  BOOL bResult;                 // results flag
  DWORD junk;                   // discard results

  hDevice = CreateFile("\\\\.\\PhysicalDrive1", // drive to open
                       0,       // don't need any access to the drive
                       FILE_SHARE_READ | FILE_SHARE_WRITE,  // share mode
                       NULL,    // default security attributes
                       OPEN_EXISTING,  // disposition
                       0,       // file attributes
                       NULL);   // don't copy any file's attributes

  if (hDevice == INVALID_HANDLE_VALUE) // we can't open the drive
  {
    return (FALSE);
  }

  bResult = DeviceIoControl(hDevice,  // device we are querying
      IOCTL_DISK_GET_DRIVE_GEOMETRY,  // operation to perform
                             NULL, 0, // no input buffer, so pass zero
                            pdg, sizeof(*pdg),  // output buffer
                            &junk, // discard count of bytes returned
                            (LPOVERLAPPED) NULL);  // synchronous I/O

  CloseHandle(hDevice);         // we're done with the handle

  return (bResult);
}

int AnalizePhysicalDrive(HANDLE &hPhysicalDrive, ULONGLONG DiskSectors)
{
	printf("Analizing PhysicalDrive (HANDLE: %ld) ...\n", hPhysicalDrive);

	// alloc buffer memory
	int BufferSize = 1024 * 1024 * 1;
	char *pBuffer = (char *)malloc(BufferSize);
	unsigned long *pBytesRead = (unsigned long *)malloc(1);

	// set file pointer, to test different physical disk areas
//	LARGE_INTEGER seek_pos;
//	LARGE_INTEGER seek_return;
//	if (!SetFilePointerEx(hPhysicalDrive, seek_pos, &seek_return, FILE_END))
//	printf("Error seeking in PhysicalDrive. Error: %ld.\n", GetLastError());

	// read first N sectors and print them out
	for (int i = 0; i < 58628; i++)
		{
			if (ReadFile(hPhysicalDrive, pBuffer, BufferSize, pBytesRead, NULL))
			{
				// read successfull, print it on the console
				if ((int)*pBytesRead != BufferSize) printf("Incorect number of bytes read at i = %d.\n", i);
				// show progress indicator
				if (!(i % 100)) printf("Progress: %d / 58628\n", i);
			}
			else
				printf("Error reading Sector(s) %i. Error %d.\n", i, GetLastError());
		}

	// clean memory
	free(pBuffer);

	return 0;
}

int main(int argc, char *argv[])
{
	
	// Global variables
  DISK_GEOMETRY pdg;            // disk drive geometry structure
  BOOL bResult;                 // generic results flag
  ULONGLONG DiskSize;           // size of the drive, in bytes
	ULONGLONG DiskSectors;				// number of sectors of the disk
	HANDLE hPhysicalDrive;				// handle to physical drive

	// Get drive geometry
  bResult = GetDriveGeometry (&pdg);

	// Analize drive geometry
  if (bResult)
  {
    printf("Cylinders = %I64d\n", pdg.Cylinders);
    printf("Tracks per cylinder = %ld\n", (ULONG) pdg.TracksPerCylinder);
    printf("Sectors per track = %ld\n", (ULONG) pdg.SectorsPerTrack);
    printf("Bytes per sector = %ld\n", (ULONG) pdg.BytesPerSector);
		DiskSectors = pdg.Cylinders.QuadPart * (ULONG)pdg.TracksPerCylinder * (ULONG)pdg.SectorsPerTrack;
    DiskSize = DiskSectors * (ULONG)pdg.BytesPerSector;
    printf("Disk size = %I64d (Bytes) = %I64d (Mb) ~= %I64d (Gb)\n", DiskSize, DiskSize / (1024 * 1024), DiskSize / (1024 * 1024 * 1024));
  }
	else
	{
		printf ("Attempt to get drive geometry failed. Error %ld.\n", GetLastError ());
  };

	// Open a handle to the physycal drive
	hPhysicalDrive = CreateFile(
			"\\\\?\\PHYSICALDRIVE1",                        // file name
			GENERIC_READ,									                  // access mode
			FILE_SHARE_READ | FILE_SHARE_WRITE,							// share mode
			NULL,																						// security descriptor
			OPEN_EXISTING,																	// how to create
			FILE_ATTRIBUTE_NORMAL,														// file attributes
			NULL																						// handle to template file
			);

	// Analize drive handle
  if (hPhysicalDrive)
  {
		printf("PhysicalDrive handle successfully opened.\n");
		// parse the handle to analizor function
		AnalizePhysicalDrive (hPhysicalDrive, DiskSectors);
  }
	else
	{
		printf("Attempt to open a handle to PhysicalDrive failed. Error %ld.\n", GetLastError ());
  };


	// Close the handle to physical drive
	if (CloseHandle(hPhysicalDrive))
		printf("PhysicalDrive handle closed successfull.\n");
	else
		printf("Attempt to close PhysicalDrive handle failed. Error %ld.\n", GetLastError ());


  return 0; // successfull program execution
}